use std::time::Duration;

use color_eyre::{Result, eyre::Ok};
use rand::seq::SliceRandom;
use ratatui::{
    DefaultTerminal, Frame,
    crossterm::event::{self, Event, KeyCode, KeyEventKind},
    layout::{Constraint, Direction, Layout},
    style::{Color, Style, Stylize},
    text::Line,
    widgets::{Bar, BarChart, BarGroup, Block},
};
use tokio::{
    sync::mpsc::{UnboundedReceiver, UnboundedSender},
    time::sleep,
};

use crate::{
    algorithm::{self},
    tui::Message,
};

pub enum SortingMessage {
    Tick,
    Finished,
    Start,
}

pub struct DrawableVec {
    vec: Vec<u32>,
    event_rx: UnboundedReceiver<SortingMessage>,
    event_tx: UnboundedSender<Message>,
    algo: String,
}

impl DrawableVec {
    pub fn new(
        event_tx: UnboundedSender<Message>,
        event_rx: UnboundedReceiver<SortingMessage>,
        vec: Vec<u32>,
    ) -> DrawableVec {
        return DrawableVec {
            vec,
            event_rx,
            event_tx: event_tx.clone(),
            algo: "insertSort".to_string(),
        };
    }

    pub async fn run(mut self) -> Result<()> {
        loop {
            let event = self.event_rx.recv().await;
            match event {
                Some(SortingMessage::Start) => break,
                _ => (),
            }
        }
        algorithm::sort(self.algo.clone(), &mut self).await?;
        self.event_tx.send(Message::SimulationFinished)?;
        Ok(())
    }

    pub fn len(&self) -> usize {
        return self.vec.len();
    }

    pub async fn swap(&mut self, a: usize, b: usize) -> Result<()> {
        self.vec.swap(a, b);
        self.event_tx.send(Message::SwapEvent(a, b))?;

        self.wait_for_next_tick().await;

        Ok(())
    }

    pub async fn set(&mut self, i: usize, elem: u32) -> Result<()> {
        self.vec[i] = elem;
        self.event_tx.send(Message::SetEvent(i, elem))?;

        self.wait_for_next_tick().await;

        Ok(())
    }

    async fn wait_for_next_tick(&mut self) {
        loop {
            match self.event_rx.recv().await {
                Some(SortingMessage::Tick) => break,
                _ => (),
            };
        }
    }

    pub fn lessThan(&self, a: usize, b: usize) -> bool {
        self.vec[a] < self.vec[b]
    }

    pub fn elements(&mut self) -> std::slice::Iter<'_, u32> {
        self.vec.iter()
    }

    pub fn get(&mut self, i: usize) -> u32 {
        *self.vec.get(i).unwrap()
    }

    pub fn randomize(&mut self) {
        self.vec.shuffle(&mut rand::rng());
    }
    pub fn isSorted(&mut self) -> bool {
        let mut prev: u32 = 0;
        for bar in self.elements() {
            if *bar < prev {
                return false;
            } else {
                prev = *bar;
            }
        }
        true
    }

    pub fn lessThanEqual(&mut self, a: usize, b: u32) -> bool {
        return self.get(a) <= b;
    }

    pub fn getListClone(&self) -> Vec<u32> {
        self.vec.clone()
    }
}
