use std::{fmt::format, time::Duration};

use crate::pages::Pages;
use crate::pages::{main::MainView, sorting::SortingView};
use crate::{
    drawableVec::{self, DrawableVec, SortingMessage},
    fps_counter::FpsCounter,
    graphics::vertical_barchart,
};
use color_eyre::{Result, eyre::Ok};
use crossterm::event::{Event, KeyCode, KeyEventKind, MouseEventKind};
use rand::{rand_core::impls, seq::SliceRandom};
use ratatui::{
    Terminal,
    layout::{Constraint, Layout, Margin},
    prelude::CrosstermBackend,
    style::Stylize,
    widgets::Block,
};
use tokio::{
    sync::mpsc::{self, UnboundedReceiver, UnboundedSender},
    task::JoinHandle,
    time,
};
pub enum Message {
    Quit,
    Tick,
    Render,
    StartSimulation,
    SwapEvent(usize, usize),
    SetEvent(usize, u32),
    SwapGui,
    SimulationFinished
}

pub struct State {
    pub vec: Vec<u32>,
}

pub struct Tui {
    terminal: Terminal<CrosstermBackend<std::io::Stdout>>,
    tick_rate: f64,
    frame_rate: f64,
    event_rx: UnboundedReceiver<Message>,
    event_tx: UnboundedSender<Message>,
    sorting_tx: Option<UnboundedSender<SortingMessage>>,
    should_exit: bool,
    fps_counter: FpsCounter,
    state: State,
    page: Pages,
}

#[derive(Clone, Debug)]
pub enum UpdateCommand {
    None,
    Quit,
}

impl Tui {
    pub fn new() -> Result<Self> {
        let terminal = ratatui::init();
        let (event_tx, event_rx): (UnboundedSender<Message>, UnboundedReceiver<Message>) =
            mpsc::unbounded_channel();
        let mut vec = (Vec::from_iter(1..50));
        vec.shuffle(&mut rand::rng());

        let page = MainView::new();

        Ok(Self {
            terminal,
            tick_rate: 30.,
            frame_rate: 60.,
            event_rx,
            event_tx,
            sorting_tx: None,
            should_exit: false,
            fps_counter: FpsCounter::new(),
            page: Pages::Main(page),
            state: State { vec },
        })
    }

    pub async fn run(&mut self) -> Result<()> {
        let tick_rate = Duration::from_secs_f64(1.0 / self.tick_rate);
        let frame_rate = Duration::from_secs_f64(1.0 / self.frame_rate);
        let mut tick_interval = time::interval(tick_rate);
        let mut frame_interval = time::interval(frame_rate);
        while !self.should_exit {
            tokio::select! {
                _tick = tick_interval.tick() => {
                    self.event_tx.send(Message::Tick)?;
                    match &self.sorting_tx {
                        Some(tx) => tx.send(SortingMessage::Tick)?,
                        None => (),
                    }
                }
                _frame = frame_interval.tick() => {
                    self.event_tx.send(Message::Render)?
                }
                Some(message) = self.event_rx.recv() => {
                    match self.update(message).await? {
                        UpdateCommand::Quit => return {
                            self.should_exit = true;
                            Ok(())
                        },
                        UpdateCommand::None => continue,
                    }
                }
                Result::Ok(ready) = tokio::task::spawn_blocking(|| crossterm::event::poll(Duration::from_millis(100))) => {
                    match ready? {
                        true => {
                            let event = crossterm::event::read()?;
                            self.handle_event(event)?;
                        }
                        false => continue,
                    }
                }
            }
        }
        Ok(())
    }

    async fn update(&mut self, message: Message) -> Result<UpdateCommand> {
        match message {
            Message::Quit => Ok(UpdateCommand::Quit),
            Message::Tick => Ok(UpdateCommand::None),
            Message::Render => {
                self.view()?;
                self.fps_counter.tick();
                Ok(UpdateCommand::None)
            }
            Message::StartSimulation => {
                match &self.sorting_tx {
                    Some(tx) => tx.send(SortingMessage::Start)?,
                    None => (),
                }
                Ok(UpdateCommand::None)
            }
            Message::SwapEvent(a, b) => {
                self.state.vec.swap(a, b);
                Ok(UpdateCommand::None)
            }
            Message::SetEvent(i, e) => {
                self.state.vec.insert(i, e);
                Ok(UpdateCommand::None)
            }
            Message::SwapGui => {
                self.page = match self.page {
                    Pages::Main(_) => {
                        let (sorting_tx, sorting_rx) = mpsc::unbounded_channel();
                        self.sorting_tx = Some(sorting_tx);
                        let drawable_vec = DrawableVec::new(
                            self.event_tx.clone(),
                            sorting_rx,
                            self.state.vec.clone(),
                        );
                        tokio::spawn(drawable_vec.run());
                        Pages::SortingView(SortingView::new())
                    }
                    Pages::SortingView(_) => Pages::Main(MainView::new()),
                };
                Ok(UpdateCommand::None)
            }
            Message::SimulationFinished => {
                self.sorting_tx = None;
                Ok(UpdateCommand::None)
            }
        }
    }

    fn view(&mut self) -> Result<()> {
        self.terminal.draw(|frame| {
            let [title, vertical] = Layout::vertical([Constraint::Length(1), Constraint::Fill(1)])
                .spacing(1)
                .areas(frame.area());
            frame.render_widget("Sorting!".bold().into_centered_line(), title);

            match &self.page {
                Pages::Main(main_view) => {
                    let _ = main_view.render(frame, &self.state, &vertical, self.fps_counter.fps);
                }
                Pages::SortingView(sorting_view) => {
                    let _ =
                        sorting_view.render(frame, &self.state, &vertical, self.fps_counter.fps);
                }
            };
        })?;

        Ok(())
    }

    fn handle_event(&self, event: Event) -> Result<()> {
        match event {
            Event::Key(key) => {
                if key.kind == KeyEventKind::Press && key.code == KeyCode::Esc {
                    self.event_tx.send(Message::Quit)?;
                } else if key.kind == KeyEventKind::Press && key.code == KeyCode::Enter {
                    match &self.sorting_tx {
                        Some(tx) => tx.send(SortingMessage::Start)?,
                        None => (),
                    };
                } else if key.kind == KeyEventKind::Press && key.code == KeyCode::Char('a') {
                    self.event_tx.send(Message::SwapGui)?;
                }
            }
            _ => {}
        }
        Ok(())
    }
}
